# type: ignore[attr-defined]
# flake8: noqa

"""
    Parea API SDK

    The Parea SDK allows you to interact with Parea from your product or service.
    To install the official [Python SDK](https://pypi.org/project/parea/),
    run the following command:  ```bash pip install parea ```.
"""
import sys
from importlib import metadata as importlib_metadata

from parea.cache import InMemoryCache, RedisCache
from parea.client import Parea
from parea.experiment.cli import experiment as _experiment_cli
from parea.experiment.dvc import parea_dvc_initialized
from parea.experiment.experiment import Experiment
from parea.helpers import date_and_time_string_to_timestamp, gen_trace_id, to_date_and_time_string, write_trace_logs_to_csv
from parea.parea_logger import parea_logger
from parea.utils.trace_integrations.langchain import PareaAILangchainTracer
from parea.utils.trace_utils import get_current_trace_id, get_root_trace_id, trace, trace_insert
from parea.wrapper.openai_raw_api_tracer import aprocess_stream_and_yield, process_stream_and_yield
from parea.wrapper.utils import convert_openai_raw_to_log


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()


def main():
    args = sys.argv[1:]
    if args[0] == "experiment":
        _experiment_cli(args[1:])
    elif args[0] == "dvc-init":
        parea_dvc_initialized(only_check=False)
    else:
        print(f"Unknown command: '{args[0]}'")


if __name__ == "__main__":
    main()



from typing import Any, Iterable, Optional

from enum import Enum

from attrs import define, field, validators

from parea.schemas.log import LLMInputs, Log


@define
class Completion:
    inference_id: Optional[str] = None
    parent_trace_id: Optional[str] = None
    root_trace_id: Optional[str] = None
    trace_name: Optional[str] = None
    llm_inputs: Optional[dict[str, Any]] = None
    llm_configuration: LLMInputs = LLMInputs()
    end_user_identifier: Optional[str] = None
    deployment_id: Optional[str] = None
    name: Optional[str] = None
    metadata: Optional[dict] = None
    tags: Optional[list[str]] = field(factory=list)
    target: Optional[str] = None
    cache: bool = True
    log_omit_inputs: bool = False
    log_omit_outputs: bool = False
    log_omit: bool = False
    experiment_uuid: Optional[str] = None
    project_uuid: str = "default"


@define
class CompletionResponse:
    inference_id: str
    content: str
    latency: float
    input_tokens: int
    output_tokens: int
    total_tokens: int
    cost: float
    model: str
    provider: str
    cache_hit: bool
    status: str
    start_timestamp: str
    end_timestamp: str
    error: Optional[str] = None


@define
class UseDeployedPrompt:
    deployment_id: str
    llm_inputs: Optional[dict[str, Any]] = None


@define
class Prompt:
    raw_messages: list[dict[str, Any]]
    messages: list[dict[str, Any]]
    inputs: Optional[dict[str, Any]] = None


@define
class UseDeployedPromptResponse:
    deployment_id: str
    name: Optional[str] = None
    functions: Optional[list[str]] = None
    function_call: Optional[str] = None
    prompt: Optional[Prompt] = None
    model: Optional[str] = None
    provider: Optional[str] = None
    model_params: Optional[dict[str, Any]] = None


@define
class FeedbackRequest:
    score: float = field(validator=[validators.ge(0), validators.le(1)])
    trace_id: Optional[str] = None
    inference_id: Optional[str] = None
    name: Optional[str] = None
    target: Optional[str] = None


@define
class NamedEvaluationScore:
    name: str
    score: float = field(validator=[validators.ge(0), validators.le(1)])


@define
class TraceLog(Log):
    trace_id: Optional[str] = field(default=None, validator=validators.instance_of(str))
    parent_trace_id: Optional[str] = field(default=None, validator=validators.instance_of(str))
    root_trace_id: Optional[str] = field(default=None, validator=validators.instance_of(str))
    start_timestamp: Optional[str] = field(default=None, validator=validators.instance_of(str))
    organization_id: Optional[str] = None
    project_uuid: Optional[str] = None

    # metrics filled from completion
    error: Optional[str] = None
    status: Optional[str] = None
    deployment_id: Optional[str] = None
    cache_hit: bool = False
    output_for_eval_metrics: Optional[str] = None
    evaluation_metric_names: Optional[list[str]] = field(factory=list)
    scores: Optional[list[NamedEvaluationScore]] = field(factory=list)
    apply_eval_frac: Optional[float] = 1.0
    feedback_score: Optional[float] = None

    # info filled from decorator
    trace_name: Optional[str] = None
    children: list[str] = field(factory=list)

    # metrics filled from either decorator or completion
    end_timestamp: Optional[str] = None
    end_user_identifier: Optional[str] = None
    metadata: Optional[dict[str, Any]] = None
    tags: Optional[list[str]] = field(factory=list)
    experiment_uuid: Optional[str] = None


@define
class TraceLogTree(TraceLog):
    children: Optional[list[TraceLog]] = field(factory=list)


@define
class CacheRequest:
    configuration: LLMInputs = LLMInputs()


@define
class UpdateLog:
    trace_id: str
    field_name_to_value_map: dict[str, Any]


@define
class CreateExperimentRequest:
    name: str


@define
class ExperimentSchema(CreateExperimentRequest):
    uuid: str
    created_at: str


@define
class EvaluationScoreSchema(NamedEvaluationScore):
    id: Optional[int] = None


@define
class TraceStatsSchema:
    trace_id: str
    latency: Optional[float] = 0.0
    input_tokens: Optional[int] = 0
    output_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0
    cost: Optional[float] = None
    scores: Optional[list[EvaluationScoreSchema]] = field(factory=list)


@define
class ExperimentStatsSchema:
    parent_trace_stats: list[TraceStatsSchema] = field(factory=list)

    def cumulative_avg_score(self) -> float:
        """Returns the average score across all evals."""
        scores = [score.score for trace_stat in self.parent_trace_stats for score in trace_stat.scores]
        return sum(scores) / len(scores) if scores else 0.0

    def avg_score(self, score_name: str) -> float:
        """Returns the average score for a given eval."""
        scores = [score.score for trace_stat in self.parent_trace_stats for score in trace_stat.scores if score.name == score_name]
        return sum(scores) / len(scores) if scores else 0.0


class UpdateTraceScenario(str, Enum):
    RESULT: str = "result"
    ERROR: str = "error"
    CHAIN: str = "chain"


@define
class CreateGetProjectSchema:
    name: str


@define
class ProjectSchema(CreateGetProjectSchema):
    uuid: str
    created_at: str


@define
class CreateGetProjectResponseSchema(ProjectSchema):
    was_created: bool


@define
class TestCase:
    id: int
    test_case_collection_id: int
    inputs: dict[str, str] = field(factory=dict)
    target: Optional[str] = None
    tags: list[str] = field(factory=list)


@define
class TestCaseCollection:
    id: int
    name: str
    created_at: str
    last_updated_at: str
    column_names: list[str] = field(factory=list)
    test_cases: dict[int, TestCase] = field(factory=dict)

    def get_all_test_case_inputs(self) -> Iterable[dict[str, str]]:
        return (test_case.inputs for test_case in self.test_cases.values())

    def num_test_cases(self) -> int:
        return len(self.test_cases)

    def get_all_test_case_targets(self) -> Iterable[str]:
        return (test_case.target for test_case in self.test_cases.values())

    def get_all_test_inputs_and_targets(self) -> Iterable[tuple[dict[str, str], Optional[str]]]:
        return ((test_case.inputs, test_case.target) for test_case in self.test_cases.values())


@define
class CreateTestCase:
    inputs: dict[str, str]
    target: Optional[str] = None
    tags: list[str] = field(factory=list)


@define
class CreateTestCaseCollection:
    name: str
    # column names excluding reserved names, target and tags
    column_names: list[str] = field(factory=list)
    test_cases: list[CreateTestCase] = field(factory=list)



from typing import Callable, Optional, Union

import asyncio
import inspect
import os
import time
from collections import defaultdict
from collections.abc import Iterable

from attrs import define, field
from tqdm import tqdm
from tqdm.asyncio import tqdm_asyncio

from parea import Parea
from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.experiment.dvc import save_results_to_dvc_if_init
from parea.helpers import gen_random_name
from parea.schemas.models import CreateExperimentRequest, ExperimentSchema, ExperimentStatsSchema
from parea.utils.trace_utils import thread_ids_running_evals
from parea.utils.universal_encoder import json_dumps

STAT_ATTRS = ["latency", "input_tokens", "output_tokens", "total_tokens", "cost"]


def calculate_avg_std_for_experiment(experiment_stats: ExperimentStatsSchema) -> dict[str, str]:
    accumulators = defaultdict(float)
    counts = defaultdict(int)
    score_accumulators = defaultdict(float)
    score_counts = defaultdict(int)

    for trace_stat in experiment_stats.parent_trace_stats:
        for attr in STAT_ATTRS:
            value = getattr(trace_stat, attr, None)
            if value is not None:
                accumulators[attr] += value
                counts[attr] += 1

        for score in trace_stat.scores:
            score_accumulators[score.name] += score.score
            score_counts[score.name] += 1

    averages = {attr: "N/A" if counts[attr] == 0 else f"{accumulators[attr] / counts[attr]:.2f}" for attr in accumulators}

    score_averages = {f"{name}": "N/A" if score_counts[name] == 0 else f"{score_accumulators[name] / score_counts[name]:.2f}" for name in score_accumulators}

    averages.update(score_averages)

    return averages


def async_wrapper(fn, **kwargs):
    return asyncio.run(fn(**kwargs))


async def experiment(name: str, data: Union[str, Iterable[dict]], func: Callable, p: Parea) -> ExperimentStatsSchema:
    """Creates an experiment and runs the function on the data iterator."""
    if isinstance(data, str):
        print(f"Fetching test collection: {data}")
        test_collection = await p.aget_collection(data)
        print(f"Fetched {test_collection.num_test_cases()} test cases from collection: {data} \n")
        data: Iterable[dict] = test_collection.get_all_test_case_inputs()

    experiment_schema: ExperimentSchema = p.create_experiment(CreateExperimentRequest(name=name))
    experiment_uuid = experiment_schema.uuid
    os.environ[PAREA_OS_ENV_EXPERIMENT_UUID] = experiment_uuid

    max_parallel_calls = 10
    sem = asyncio.Semaphore(max_parallel_calls)

    async def limit_concurrency(data_input):
        async with sem:
            return await func(**data_input)

    if inspect.iscoroutinefunction(func):
        tasks = [limit_concurrency(data_input) for data_input in data]
        for result in tqdm_asyncio(tasks):
            await result
    else:
        for data_input in tqdm(data):
            func(**data_input)

    total_evals = len(thread_ids_running_evals.get())
    with tqdm(total=total_evals, dynamic_ncols=True) as pbar:
        while thread_ids_running_evals.get():
            pbar.set_description(f"Waiting for evaluations to finish")
            pbar.update(total_evals - len(thread_ids_running_evals.get()))
            total_evals = len(thread_ids_running_evals.get())
            await asyncio.sleep(0.5)

    time.sleep(2)
    experiment_stats: ExperimentStatsSchema = p.finish_experiment(experiment_uuid)
    stat_name_to_avg_std = calculate_avg_std_for_experiment(experiment_stats)
    print(f"Experiment {name} stats:\n{json_dumps(stat_name_to_avg_std, indent=2)}\n\n")
    print(f"View experiment & traces at: https://app.parea.ai/experiments/{experiment_uuid}\n")
    save_results_to_dvc_if_init(name, stat_name_to_avg_std)
    return experiment_stats


_experiments = []


@define
class Experiment:
    # If your dataset is defined locally it should be an iterable of k/v
    # pairs matching the expected inputs of your function. To reference a test collection you
    # have saved on Parea, use the collection name as a string.
    data: Union[str, Iterable[dict]]
    # The function to run. This function should accept inputs that match the keys of the data field.
    func: Callable = field()
    experiment_stats: ExperimentStatsSchema = field(init=False, default=None)
    p: Parea = field(default=None)
    name: str = field(init=False)

    def __attrs_post_init__(self):
        global _experiments
        _experiments.append(self)

    def _gen_name_if_none(self, name: Optional[str]):
        if not name:
            self.name = gen_random_name()
            print(f"Experiment name set to: {self.name}, since a name was not provided.")
        else:
            self.name = name

    def run(self, name: Optional[str] = None) -> None:
        """Run the experiment and save the results to DVC.
        param name: The name of the experiment. This name must be unique across experiment runs.
        If no name is provided a memorable name will be generated automatically.
        """
        self._gen_name_if_none(name)
        self.experiment_stats = asyncio.run(experiment(self.name, self.data, self.func, self.p))



import asyncio
import os

from dotenv import load_dotenv
from openai.lib.azure import AsyncAzureOpenAI, AzureOpenAI

from parea import Parea, trace
from parea.cookbook.data.openai_input_examples import functions_example, simple_example

load_dotenv()

client = AzureOpenAI(
    api_version="2023-12-01-preview",
    api_key=os.getenv("AZURE_OAI_API_KEY"),
    azure_endpoint=os.getenv("AZURE_OAI_ENDPOINT"),
)
aclient = AsyncAzureOpenAI(
    api_version="2023-12-01-preview",
    api_key=os.getenv("AZURE_OAI_API_KEY"),
    azure_endpoint=os.getenv("AZURE_OAI_ENDPOINT"),
)

p = Parea(api_key=os.getenv("PAREA_API_KEY"))
p.wrap_openai_client(client)
p.wrap_openai_client(aclient)


@trace
def call_azure(data: dict):
    response = client.chat.completions.create(**data)
    print(response)


@trace
def call_azure_stream(data: dict):
    data["stream"] = True
    stream = client.chat.completions.create(**data)
    for chunk in stream:
        if chunk.choices:
            print(chunk.choices[0].delta or "")


@trace
async def acall_azure(data: dict):
    response = await aclient.chat.completions.create(**data)
    print(response)


@trace
async def acall_azure_stream(data: dict):
    data["stream"] = True
    stream = await aclient.chat.completions.create(**data)
    async for chunk in stream:
        if chunk.choices:
            print(chunk.choices[0].delta or "")


if __name__ == "__main__":
    azure_model = "AZURE_MODEL_NAME"  # replace with your model name
    functions_example["model"] = azure_model
    simple_example["model"] = azure_model
    call_azure(functions_example)
    # call_azure_stream(simple_example)
    # call_azure_stream(functions_example)
    asyncio.run(acall_azure(simple_example))
    # asyncio.run(acall_azure(functions_example))
    # asyncio.run(acall_azure_stream(simple_example))
    asyncio.run(acall_azure_stream(functions_example))



tool_calling_example = {
    "model": "gpt-3.5-turbo-0125",
    "messages": [{"role": "user", "content": "What's the weather like in San Francisco, Tokyo, and Paris?"}],
    "tools": [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather in a given location",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                    },
                    "required": ["location"],
                },
            },
        }
    ],
    "tool_choice": "auto",
}

functions_example = {
    "model": "gpt-3.5-turbo-0125",
    "messages": [
        {
            "role": "system",
            "content": f"You are a sophisticated AI assistant, "
            f"a specialist in user intent detection and interpretation. "
            f"Your task is to perceive and respond to the user's needs, even when they're expressed "
            f"in an indirect or direct manner. You excel in recognizing subtle cues: for example, "
            f"if a user states they are 'hungry', you should assume they are seeking nearby dining "
            f"options such as a restaurant or a cafe. If they indicate feeling 'tired', 'weary', "
            f"or mention a long journey, interpret this as a request for accommodation options like "
            f"hotels or guest houses. However, remember to navigate the fine line of interpretation "
            f"and assumption: if a user's intent is unclear or can be interpreted in multiple ways, "
            f"do not hesitate to politely ask for additional clarification. Make sure to tailor your "
            f"responses to the user based on their preferences and past experiences which can "
            f"be found here: Name: John Doe",
        },
        {"role": "user", "content": "I'm hungry"},
    ],
    "functions": [
        {
            "name": "call_google_places_api",
            "description": f"""
            This function calls the Google Places API to find the top places of a specified type near
            a specific location. It can be used when a user expresses a need (e.g., feeling hungry or tired) or wants to
            find a certain type of place (e.g., restaurant or hotel).
        """,
            "parameters": {"type": "object", "properties": {"place_type": {"type": "string", "description": "The type of place to search for."}}},
            "result": {"type": "array", "items": {"type": "string"}},
        }
    ],
}

simple_example = {
    "model": "gpt-3.5-turbo-0125",
    "messages": [{"role": "system", "content": "You are a helpful assistant."}, {"role": "user", "content": "Hello!"}],
}

simple_example_json = {
    "model": "gpt-3.5-turbo-0125",
    "messages": [{"role": "system", "content": "You are a helpful assistant talking JSON."}, {"role": "user", "content": "Hello!"}],
    "response_format": {"type": "json_object"},
}



import os

import boto3
from dotenv import load_dotenv
from langchain.document_loaders import TextLoader
from langchain.llms.bedrock import Bedrock
from langchain.output_parsers import XMLOutputParser
from langchain.prompts import PromptTemplate
from langchain.schema.output_parser import StrOutputParser
from langchain.text_splitter import RecursiveCharacterTextSplitter

from parea import Parea, PareaAILangchainTracer

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))
handler = PareaAILangchainTracer()


def get_docs():
    loader = TextLoader("../data/2022-letter.txt")
    letter = loader.load()
    text_splitter = RecursiveCharacterTextSplitter(separators=["\n\n", "\n"], chunk_size=4000, chunk_overlap=100)
    return text_splitter.split_documents(letter)


xml_parser = XMLOutputParser(tags=["insight"])
str_parser = StrOutputParser()

insight_prompt = PromptTemplate(
    template="""

    Human:
    {instructions} : \"{document}\"
    Format help: {format_instructions}.
    Assistant:""",
    input_variables=["instructions", "document"],
    partial_variables={"format_instructions": xml_parser.get_format_instructions()},
)

summary_prompt = PromptTemplate(
    template="""

    Human:
    {instructions} : \"{document}\"
    Assistant:""",
    input_variables=["instructions", "document"],
)

docs = get_docs()
bedrock_client = boto3.client("bedrock-runtime", region_name="us-east-1")
bedrock_llm = Bedrock(
    client=bedrock_client,
    model_id="amazon.titan-text-express-v1",
    model_kwargs={"maxTokenCount": 4096, "stopSequences": [], "temperature": 0, "topP": 1},
)

insight_chain = insight_prompt | bedrock_llm | StrOutputParser()
summary_chain = summary_prompt | bedrock_llm | StrOutputParser()


def get_insights(docs):
    insights = []
    for i in range(len(docs)):
        insight = insight_chain.invoke(
            {"instructions": "Provide Key insights from the following text", "document": {docs[i].page_content}}, config={"callbacks": [PareaAILangchainTracer()]}
        )
        insights.append(insight)
    return insights


def main():
    print("Starting")
    insights = get_insights(docs)
    print(insights)
    summary = summary_chain.invoke(
        {
            "instructions": "You will be provided with multiple sets of insights. Compile and summarize these "
            "insights and provide key takeaways in one concise paragraph. Do not use the original xml "
            "tags. Just provide a paragraph with your compiled insights.",
            "document": {"\n".join(insights)},
        },
        config={"callbacks": [PareaAILangchainTracer()]},
    )
    print(summary)
    print("Done")


if __name__ == "__main__":
    main()



import os

from dotenv import load_dotenv

from parea import Parea, trace
from parea.schemas import Completion, LLMInputs, Log, Message, ModelParams, Role

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))


def eval_func(log: Log) -> float:
    from random import random
    from time import sleep

    sleep(random() * 10)
    return random()


# annotate the function with the trace decorator and pass the evaluation function(s)
@trace(eval_funcs=[eval_func])
def func(lang: str, framework: str) -> str:
    return p.completion(
        data=Completion(
            llm_configuration=LLMInputs(
                model="gpt-3.5-turbo",
                model_params=ModelParams(temp=1),
                messages=[
                    Message(role=Role.user, content=f"Write a hello world program in {lang} using {framework}"),
                ],
            )
        )
    ).content


if __name__ == "__main__":
    p.experiment(
        data="Hello World Example",  # this is the name of my Test Collection in Parea (TestHub page)
        func=func,
    ).run(name="hello-world-example")



from typing import Any

import json
import os

from attrs import asdict, define, field

from parea.api_client import HTTPClient
from parea.cache.redis import RedisCache
from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.schemas.log import TraceIntegrations
from parea.schemas.models import TraceLog, UpdateLog
from parea.utils.universal_encoder import json_dumps

LOG_ENDPOINT = "/trace_log"
VENDOR_LOG_ENDPOINT = "/trace_log/{vendor}"


@define
class PareaLogger:
    _client: HTTPClient = field(init=False, default=None)
    _redis_cache: RedisCache = field(init=False, default=None)
    _project_uuid: str = field(init=False, default=None)

    def set_client(self, client: HTTPClient) -> None:
        self._client = client

    def set_redis_cache(self, cache: RedisCache) -> None:
        self._redis_cache = cache

    def set_project_uuid(self, project_uuid: str) -> None:
        self._project_uuid = project_uuid

    def update_log(self, data: UpdateLog) -> None:
        self._client.request(
            "PUT",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    def record_log(self, data: TraceLog) -> None:
        data.project_uuid = self._project_uuid
        self._client.request(
            "POST",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    async def arecord_log(self, data: TraceLog) -> None:
        data.project_uuid = self._project_uuid
        await self._client.request_async(
            "POST",
            LOG_ENDPOINT,
            data=asdict(data),
        )

    def write_log(self, data: TraceLog) -> None:
        self._redis_cache.log(data)

    def default_log(self, data: TraceLog) -> None:
        if self._redis_cache:
            self.write_log(data)
        if self._client:
            self.record_log(data)

    def record_vendor_log(self, data: dict[str, Any], vendor: TraceIntegrations) -> None:
        data["project_uuid"] = self._project_uuid
        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data["experiment_uuid"] = experiment_uuid
        self._client.request(
            "POST",
            VENDOR_LOG_ENDPOINT.format(vendor=vendor.value),
            data=json.loads(json_dumps(data)),  # uuid is not serializable
        )

    async def arecord_vendor_log(self, data: dict[str, Any], vendor: TraceIntegrations) -> None:
        data["project_uuid"] = self._project_uuid
        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data["experiment_uuid"] = experiment_uuid
        await self._client.request_async(
            "POST",
            VENDOR_LOG_ENDPOINT.format(vendor=vendor.value),
            data=json.loads(json_dumps(data)),  # uuid is not serializable
        )


parea_logger = PareaLogger()



import asyncio
import os

from dotenv import load_dotenv
from openai import AsyncOpenAI, OpenAI

from parea import Parea, trace
from parea.cookbook.data.openai_input_examples import functions_example, simple_example_json

load_dotenv()

client = OpenAI(api_key=os.getenv("OPENAI_API_KEY"))
aclient = AsyncOpenAI(api_key=os.getenv("OPENAI_API_KEY"))

p = Parea(api_key=os.getenv("PAREA_API_KEY"))
p.wrap_openai_client(client)
p.wrap_openai_client(aclient)


@trace
def call_openai_stream(data: dict):
    data["stream"] = True
    stream = client.chat.completions.create(**data)
    for chunk in stream:
        print(chunk.choices[0].delta or "")


@trace
async def acall_openai_stream(data: dict):
    data["stream"] = True
    stream = await aclient.chat.completions.create(**data)
    async for chunk in stream:
        print(chunk.choices[0].delta or "")


if __name__ == "__main__":
    # call_openai_stream(simple_example)
    call_openai_stream(simple_example_json)
    # call_openai_stream(functions_example)
    # asyncio.run(acall_openai_stream(simple_example))
    asyncio.run(acall_openai_stream(functions_example))



from typing import Callable, Optional, Union

import json
import sys
from functools import lru_cache, wraps

import tiktoken
from openai import __version__ as openai_version

from parea.constants import AZURE_MODEL_INFO, OPENAI_MODEL_INFO

if openai_version.startswith("1."):
    from openai.types.chat import ChatCompletion

from parea.parea_logger import parea_logger
from parea.schemas.log import LLMInputs, ModelParams
from parea.schemas.models import UpdateLog
from parea.utils.trace_utils import get_current_trace_id, log_in_thread, trace_insert
from parea.utils.universal_encoder import json_dumps


# https://gist.github.com/JettJones/c236494013f22723c1822126df944b12
def skip_decorator_if_func_in_stack(*funcs_to_check: Callable) -> Callable:
    def decorator_wrapper(decorator: Callable) -> Callable:
        def new_decorator(self, func: Callable) -> Callable:  # Include self
            @wraps(func)
            def wrapper(*args, **kwargs):
                frame = sys._getframe().f_back
                caller_names = ""
                while frame:
                    caller_names += frame.f_code.co_name + "|"
                    frame = frame.f_back
                if any(func_to_check.__name__ in caller_names for func_to_check in funcs_to_check):
                    return func(*args, **kwargs)
                return decorator(self, func)(*args, **kwargs)  # Include self

            return wrapper

        return new_decorator

    return decorator_wrapper


def _num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613", is_azure: bool = False):
    """Return the number of tokens used by a list of messages.
    source: https://cookbook.openai.com/examples/how_to_count_tokens_with_tiktoken
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if (
        model
        in {
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-1106",
            "gpt-3.5-turbo-0125",
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
            "gpt-4-turbo-preview",
            "gpt-4-1106-preview",
            "gpt-4-0125-preview",
        }
        or is_azure
    ):
        tokens_per_message = 3
        tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-4" in model:
        print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
        return _num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}.
            See https://github.com/openai/openai-python/blob/main/chatml.md for
            information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def _num_tokens_from_functions(functions, function_call, model="gpt-3.5-turbo-0613"):
    """Return the number of tokens used by a list of functions. modified from source:
    https://community.openai.com/t/how-to-calculate-the-tokens-when-using-function-call/266573/10 tested:
    https://deepnote.com/workspace/joel-alexander-7514-38647c2d-ba7d-4573-b016-5233b373912e/project/BloomreachDemo
    -66683e80-4414-4b43-ac70-e7eabd3743c1/notebook/token_counting_openai-992f53bba3224221901dcb88befa2051
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 3 if model in ["gpt-3.5-turbo-1106", "gpt-4-1106-preview"] else 0
    for function in functions:
        function_tokens = len(encoding.encode(function.get("name", "")))
        function_tokens += len(encoding.encode(function.get("description", "")))

        if "parameters" in function:
            parameters = function["parameters"]
            if "properties" in parameters:
                for propertiesKey in parameters["properties"]:
                    function_tokens += len(encoding.encode(propertiesKey))
                    v = parameters["properties"][propertiesKey]
                    for field in v:
                        if field == "type":
                            function_tokens += 2
                            function_tokens += len(encoding.encode(v["type"]))
                        elif field == "description":
                            function_tokens += 2
                            function_tokens += len(encoding.encode(v["description"]))
                        elif field == "enum":
                            function_tokens -= 3
                            for o in v["enum"]:
                                function_tokens += 3
                                function_tokens += len(encoding.encode(o))
                        else:
                            print(f"Warning: not supported field {field}")
                function_tokens += 11

        num_tokens += function_tokens

    num_tokens += 10
    function_call_tokens = len(encoding.encode("auto")) - 1
    if isinstance(function_call, dict):
        function_call_tokens = len(encoding.encode(json_dumps(function_call))) - 1
    return num_tokens + function_call_tokens


def _num_tokens_from_string(string: str, model_name: str = "gpt-3.5-turbo") -> int:
    """Returns the number of tokens in a text string."""
    try:
        encoding = tiktoken.encoding_for_model(model_name)
    except KeyError:
        print(f"Warning: model {model_name} not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    num_tokens = len(encoding.encode(string))
    return num_tokens


def _calculate_input_tokens(
    messages: Optional[list[dict[str, str]]],
    functions: list[dict[str, str]],
    function_call: Union[str, dict[str, str]],
    model: str,
) -> int:
    is_azure = model.startswith("azure_") or model in AZURE_MODEL_INFO
    num_function_tokens = _num_tokens_from_functions(functions, function_call, model)
    num_input_tokens = _num_tokens_from_string(json_dumps(messages), model) if model == "gpt-4-vision-preview" else _num_tokens_from_messages(messages, model, is_azure)
    return num_input_tokens + num_function_tokens


def _format_function_call(response_message) -> str:
    def clean_json_string(s):
        """If OpenAI responds with improper newlines and multiple quotes, this will clean it up"""
        return json_dumps(s.replace("'", '"').replace("\\n", "\\\\n"))

    func_obj = response_message.tool_calls
    if response_message.function_call and response_message.function_call.name:
        func_obj = response_message.function_call

    calls = []
    if not isinstance(func_obj, list):
        func_obj = [func_obj]

    for call in func_obj:
        if call:
            body = getattr(call, "function", None) or call
            function_name = body.name
            try:
                function_args = json.loads(body.arguments)
            except json.decoder.JSONDecodeError:
                function_args = json.loads(clean_json_string(body.arguments))
            calls.append({"name": function_name, "arguments": function_args})
    return json_dumps(calls, indent=4)


def _kwargs_to_llm_configuration(kwargs, model=None):
    functions = kwargs.get("functions", None) or [d["function"] for d in kwargs.get("tools", [])]
    function_call_default = "auto" if functions else None
    return LLMInputs(
        model=model or kwargs.get("model", None),
        provider="openai",
        messages=kwargs.get("messages", None),
        functions=functions,
        function_call=kwargs.get("function_call", function_call_default) or kwargs.get("tool_choice", function_call_default),
        model_params=ModelParams(
            temp=kwargs.get("temperature", 1.0),
            max_length=kwargs.get("max_tokens", None),
            top_p=kwargs.get("top_p", 1.0),
            frequency_penalty=kwargs.get("frequency_penalty", 0.0),
            presence_penalty=kwargs.get("presence_penalty", 0.0),
            response_format=kwargs.get("response_format", None),
        ),
    )


@lru_cache(maxsize=128)
def _compute_cost(prompt_tokens: int, completion_tokens: int, model: str) -> float:
    if model in AZURE_MODEL_INFO:
        cost_per_token = AZURE_MODEL_INFO[model]
    else:
        cost_per_token = OPENAI_MODEL_INFO.get(model, {"prompt": 0, "completion": 0})
    cost = ((prompt_tokens * cost_per_token["prompt"]) + (completion_tokens * cost_per_token["completion"])) / 1_000_000
    cost = round(cost, 10)
    return cost


def _process_response(response, model_inputs, trace_id):
    response_message = response.choices[0].message
    if response_message.content:
        completion = response_message.content.strip()
    else:
        completion = _format_function_call(response_message)

    usage = response.usage
    trace_insert(
        {
            "configuration": _kwargs_to_llm_configuration(model_inputs, response.model),
            "output": completion,
            "input_tokens": usage.prompt_tokens,
            "output_tokens": usage.completion_tokens,
            "total_tokens": usage.prompt_tokens + usage.completion_tokens,
            "cost": _compute_cost(usage.prompt_tokens, usage.completion_tokens, response.model),
        },
        trace_id,
    )


def _process_stream_response(content: list, tools: dict, data: dict, trace_id: str):
    # format to Log schema and send to logger
    model = data.get("model")
    final_content = "".join(content)

    for tool in tools.values():
        tool["function"]["arguments"] = "".join(tool["function"]["arguments"])

    tool_calls = [t["function"] for t in tools.values()]
    for tool in tool_calls:
        tool["arguments"] = json.loads(tool["arguments"])

    completion = final_content or json_dumps(tool_calls, indent=4)

    prompt_tokens = _calculate_input_tokens(
        data.get("messages", []),
        data.get("functions", []) or [d["function"] for d in data.get("tools", [])],
        data.get("function_call", "auto") or data.get("tool_choice", "auto"),
        data.get("model"),
    )
    completion_tokens = _num_tokens_from_string(final_content if final_content else json_dumps(tool_calls), model)
    parea_logger.update_log(
        UpdateLog(
            trace_id=trace_id,
            field_name_to_value_map={
                "configuration": _kwargs_to_llm_configuration(data, model),
                "output": completion,
                "input_tokens": prompt_tokens,
                "output_tokens": completion_tokens,
                "total_tokens": prompt_tokens + completion_tokens,
                "cost": _compute_cost(prompt_tokens, completion_tokens, model),
            },
        )
    )


def convert_openai_raw_stream_to_log(content: list, tools: dict, data: dict, trace_id: str):
    log_in_thread(_process_stream_response, {"content": content, "tools": tools, "data": data, "trace_id": trace_id})


def convert_openai_raw_to_log(r: dict, data: dict):
    log_in_thread(_process_response, {"response": ChatCompletion(**r), "model_inputs": data, "trace_id": get_current_trace_id()})



import os

import bs4
from dotenv import load_dotenv
from langchain import hub
from langchain.document_loaders import WebBaseLoader
from langchain.schema import StrOutputParser
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import Chroma
from langchain_core.runnables import RunnablePassthrough
from langchain_openai import ChatOpenAI, OpenAIEmbeddings

from parea import Parea, PareaAILangchainTracer

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))

loader = WebBaseLoader(
    web_paths=("https://lilianweng.github.io/posts/2023-06-23-agent/",),
    bs_kwargs=dict(parse_only=bs4.SoupStrainer(class_=("post-content", "post-title", "post-header"))),
)
docs = loader.load()

text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=200)
splits = text_splitter.split_documents(docs)

vectorstore = Chroma.from_documents(documents=splits, embedding=OpenAIEmbeddings())
retriever = vectorstore.as_retriever()

prompt = hub.pull("rlm/rag-prompt")
llm = ChatOpenAI(temperature=0)


def format_docs(docs):
    return "\n\n".join(doc.page_content for doc in docs)


rag_chain = {"context": retriever | format_docs, "question": RunnablePassthrough()} | prompt | llm | StrOutputParser()


def main():
    response = rag_chain.invoke("What is Task Decomposition?", config={"callbacks": [PareaAILangchainTracer()]})
    print(response)


if __name__ == "__main__":
    main()



import os

from dotenv import load_dotenv

from parea import Parea, trace
from parea.evals import call_openai
from parea.schemas import Log

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))


# Evaluation function(s)
def is_between_1_and_n(log: Log) -> float:
    """Evaluates if the number is between 1 and n"""
    n = log.inputs["n"]
    try:
        return 1.0 if 1.0 <= float(log.output) <= float(n) else 0.0
    except ValueError:
        return 0.0


# annotate the function with the trace decorator and pass the evaluation function(s)
@trace(eval_funcs=[is_between_1_and_n])
def generate_random_number(n: str) -> str:
    return call_openai(
        [
            {"role": "user", "content": f"Generate a number between 1 and {n}."},
        ],
        model="gpt-3.5-turbo",
    )


# Define the experiment
# You can use the CLI command "parea experiment parea/cookbook/run_experiment.py" to execute this experiment
# or call `.run()`
# p.experiment(
#     data=[{"n": "10"}],  # Data to run the experiment on (list of dicts)
#     func=generate_random_number,  # Function to run (callable)
# )

# You can optionally run the experiment manually by calling `.run()`
if __name__ == "__main__":
    p.experiment(
        data=[{"n": "10"}],
        func=generate_random_number,
    ).run()



import os
from datetime import datetime
from operator import itemgetter

import boto3
from dotenv import load_dotenv

# LangChain libs
from langchain.chat_models import ChatOpenAI
from langchain.document_loaders import RecursiveUrlLoader
from langchain.document_transformers import Html2TextTransformer
from langchain.embeddings import OpenAIEmbeddings
from langchain.prompts import ChatPromptTemplate
from langchain.schema.output_parser import StrOutputParser
from langchain.text_splitter import TokenTextSplitter
from langchain.vectorstores import Chroma
from langchain_community.llms.bedrock import Bedrock

# Parea libs
from parea import Parea, PareaAILangchainTracer
from parea.evals import EvalFuncTuple, run_evals_in_thread_and_log
from parea.evals.general import answer_matches_target_llm_grader_factory
from parea.evals.rag import (
    answer_context_faithfulness_binary_factory,
    answer_context_faithfulness_statement_level_factory,
    context_query_relevancy_factory,
    percent_target_supported_by_context_factory,
)
from parea.schemas import LLMInputs, Log

load_dotenv()

# Need to instantiate Parea for tracing and evals
p = Parea(api_key=os.getenv("PAREA_API_KEY"))


bedrock_client = boto3.client("bedrock-runtime", region_name="us-east-1")
bedrock_model = Bedrock(client=bedrock_client, model_id="amazon.titan-text-express-v1", model_kwargs={"maxTokenCount": 4096, "stopSequences": [], "temperature": 0, "topP": 1})
bedrock = {"model_name": "amazon.titan-text-express-v1", "model": bedrock_model, "provider": "bedrock"}
openai_model = ChatOpenAI(model="gpt-3.5-turbo-16k", temperature=0)
openai = {"model_name": "gpt-3.5-turbo-16k", "model": openai_model, "provider": "openai"}


class DocumentRetriever:
    def __init__(self, url: str):
        api_loader = RecursiveUrlLoader(url)
        raw_documents = api_loader.load()

        # Transformer
        doc_transformer = Html2TextTransformer()
        transformed = doc_transformer.transform_documents(raw_documents)

        # Splitter
        text_splitter = TokenTextSplitter(
            model_name="gpt-3.5-turbo",
            chunk_size=2000,
            chunk_overlap=200,
        )
        documents = text_splitter.split_documents(transformed)

        # Define vector store based
        embeddings = OpenAIEmbeddings()
        vectorstore = Chroma.from_documents(documents, embeddings)
        self.retriever = vectorstore.as_retriever(search_kwargs={"k": 4})

    def get_retriever(self):
        return self.retriever


class DocumentationChain:
    def __init__(self, retriever, model, is_bedrock=False):
        self.is_bedrock = is_bedrock
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a helpful documentation Q&A assistant, trained to answer questions from the provided context."
                    "\nThe current time is {time}.\n\nRelevant documents will be retrieved in the following messages.",
                ),
                ("system", "{context}"),
                ("human", "{question}"),
            ]
        ).partial(time=str(datetime.now()))

        response_generator = prompt | model | StrOutputParser()

        self.chain = (
            # The runnable map here routes the original inputs to a context and a question dictionary to pass to the response generator
            {"context": itemgetter("question") | retriever | self._format_docs, "question": itemgetter("question")}
            | response_generator
        )

    def get_context(self) -> str:
        """Helper to get the context from a retrieval chain, so we can use it for evaluation metrics."""
        return self.context

    def _format_docs(self, docs) -> str:
        if self.is_bedrock:
            docs = docs[0::2]  # avoid context window limit, get every 2nd doc
        context = "\n\n".join(doc.page_content for doc in docs)
        # set context as an attribute, so we can access it later
        self.context = context
        return context

    def get_chain(self):
        return self.chain


# EXAMPLE EVALUATION TEST CASES
eval_questions = [
    "What is the population of New York City as of 2020?",
    "Which borough of New York City has the highest population? Only respond with the name of the borough.",
    "What is the economic significance of New York City?",
    "How did New York City get its name?",
    "What is the significance of the Statue of Liberty in New York City?",
]

eval_answers = [
    "8,804,190",
    "Brooklyn",
    """New York City's economic significance is vast, as it serves as the global financial capital, housing Wall
    Street and major financial institutions. Its diverse economy spans technology, media, healthcare, education,
    and more, making it resilient to economic fluctuations. NYC is a hub for international business, attracting
    global companies, and boasts a large, skilled labor force. Its real estate market, tourism, cultural industries,
    and educational institutions further fuel its economic prowess. The city's transportation network and global
    influence amplify its impact on the world stage, solidifying its status as a vital economic player and cultural
    epicenter.""",
    """New York City got its name when it came under British control in 1664. King Charles II of England granted the
    lands to his brother, the Duke of York, who named the city New York in his own honor.""",
    """The Statue of Liberty in New York City holds great significance as a symbol of the United States and its
    ideals of liberty and peace. It greeted millions of immigrants who arrived in the U.S. by ship in the late 19th
    and early 20th centuries, representing hope and freedom for those seeking a better life. It has since become an
    iconic landmark and a global symbol of cultural diversity and freedom.""",
]

# set up evaluation functions we want to test, provide the name of the relevant fields needed for each eval function
EVALS = [
    # question field only
    EvalFuncTuple(name="matches_target", func=answer_matches_target_llm_grader_factory(question_field="question")),
    # questions field and single context field
    EvalFuncTuple(name="faithfulness_binary", func=answer_context_faithfulness_binary_factory(question_field="question", context_field="context")),
    # questions field and accepts multiple context fields
    EvalFuncTuple(name="faithfulness_statement", func=answer_context_faithfulness_statement_level_factory(question_field="question", context_fields=["context"])),
    EvalFuncTuple(name="relevancy", func=context_query_relevancy_factory(question_field="question", context_fields=["context"])),
    EvalFuncTuple(name="supported_by_context", func=percent_target_supported_by_context_factory(question_field="question", context_fields=["context"])),
]


def create_log(model, question, context, output, target, provider="openai") -> Log:
    """Creates a log object for evaluation metric functions."""
    inputs = {"question": question, "context": context}  # fields named question and context to align w/ eval functions
    log_config = LLMInputs(model=model, provider=provider)  # provider and model needed for precision eval metric
    return Log(configuration=log_config, inputs=inputs, output=output, target=target)


def main():
    model = openai["model"]
    model_name = openai["model_name"]
    provider = openai["provider"]

    # instantiate tracer integration
    handler = PareaAILangchainTracer()
    # set up retriever
    retriever = DocumentRetriever("https://en.wikipedia.org/wiki/New_York_City").get_retriever()
    # set up chain
    dc = DocumentationChain(retriever, model, is_bedrock=provider == "bedrock")

    # iterate through questions and answers
    for question, answer in zip(eval_questions, eval_answers):
        # call chain and attached parea tracer as a callback for logs
        output = dc.get_chain().invoke({"question": question}, config={"callbacks": [handler]})
        print(f"Question: {question}\nAnswer: {output} \n")

        # get parent trace id from the tracer
        parent_trace_id = handler.get_parent_trace_id()
        # after chain is called, get the context for evaluation metric functions
        context = dc.get_context()
        # build log component needed for evaluation metric functions
        log = create_log(model_name, question, context, output, answer, provider)

        # helper function to run evaluation metrics in a thread to avoid blocking return of chain
        run_evals_in_thread_and_log(trace_id=str(parent_trace_id), log=log, eval_funcs=EVALS, verbose=True)


if __name__ == "__main__":
    print("Running evals...")
    main()
    print("Done!")



import os

from dotenv import load_dotenv
from langchain.agents.agent_toolkits import create_conversational_retrieval_agent, create_retriever_tool
from langchain.chat_models import ChatOpenAI
from langchain.document_loaders import TextLoader
from langchain.embeddings import OpenAIEmbeddings
from langchain.text_splitter import CharacterTextSplitter
from langchain.vectorstores import FAISS

from parea import Parea, PareaAILangchainTracer

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))

loader = TextLoader("../data/state_of_the_union.txt")


documents = loader.load()
text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
texts = text_splitter.split_documents(documents)
embeddings = OpenAIEmbeddings()
db = FAISS.from_documents(texts, embeddings)
retriever = db.as_retriever()
tool = create_retriever_tool(
    retriever,
    "search_state_of_union",
    "Searches and returns documents regarding the state-of-the-union.",
)
tools = [tool]


llm = ChatOpenAI(temperature=0)

agent_executor = create_conversational_retrieval_agent(llm, tools)


def main():
    result = agent_executor({"input": "what did the president say about kentaji brown jackson in the most recent state of the union?"}, callbacks=[PareaAILangchainTracer()])
    print(result)


if __name__ == "__main__":
    main()



from typing import Any, Callable, Optional

import contextvars
import inspect
import json
import logging
import os
import threading
import time
from collections import ChainMap
from collections.abc import AsyncGenerator, AsyncIterator, Generator, Iterator
from functools import wraps
from random import random

from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.helpers import gen_trace_id, to_date_and_time_string
from parea.parea_logger import parea_logger
from parea.schemas.models import NamedEvaluationScore, TraceLog, UpdateLog, UpdateTraceScenario
from parea.utils.universal_encoder import json_dumps

logger = logging.getLogger()


# Context variable to maintain the current trace context stack
trace_context = contextvars.ContextVar("trace_context", default=[])

# A dictionary to hold trace data for each trace
trace_data = contextvars.ContextVar("trace_data", default={})

# Context variable to maintain running evals in thread
thread_ids_running_evals = contextvars.ContextVar("thread_ids_running_evals", default=[])


def log_in_thread(target_func: Callable, data: dict[str, Any]):
    logging_thread = threading.Thread(target=target_func, kwargs=data)
    logging_thread.start()


def merge(old, new):
    if isinstance(old, dict) and isinstance(new, dict):
        return dict(ChainMap(new, old))
    if isinstance(old, list) and isinstance(new, list):
        return old + new
    return new


def check_multiple_return_values(func) -> bool:
    specs = inspect.getfullargspec(func)
    try:
        r = specs.annotations.get("return")
        if r and r.__origin__ == tuple:
            return len(r.__args__) > 1
    except Exception:
        return False


def make_output(result, islist) -> str:
    if islist:
        json_list = [json_dumps(r) for r in result]
        return json_dumps(json_list)
    elif isinstance(result, str):
        return result
    else:
        return json_dumps(result)


def get_current_trace_id() -> str:
    stack = trace_context.get()
    if stack:
        return stack[-1]
    return ""


def get_root_trace_id() -> str:
    stack = trace_context.get()
    if stack:
        return stack[0]
    return ""


def trace_insert(data: dict[str, Any], trace_id: Optional[str] = None):
    current_trace_id = trace_id or get_current_trace_id()
    current_trace_data: TraceLog = trace_data.get()[current_trace_id]
    if not current_trace_data:
        return
    for key, new_value in data.items():
        existing_value = current_trace_data.__getattribute__(key)
        current_trace_data.__setattr__(key, merge(existing_value, new_value) if existing_value else new_value)


def fill_trace_data(trace_id: str, data: dict[str, Any], scenario: UpdateTraceScenario):
    if scenario == UpdateTraceScenario.RESULT:
        if not isinstance(data["result"], (Generator, AsyncGenerator, AsyncIterator, Iterator)):
            trace_data.get()[trace_id].output = make_output(data["result"], data.get("output_as_list", False))
        trace_data.get()[trace_id].status = "success"
        trace_data.get()[trace_id].evaluation_metric_names = data.get("eval_funcs_names")
    elif scenario == UpdateTraceScenario.ERROR:
        trace_data.get()[trace_id].error = data["error"]
        trace_data.get()[trace_id].status = "error"
    elif scenario == UpdateTraceScenario.CHAIN:
        trace_data.get()[trace_id].parent_trace_id = data["parent_trace_id"]
        trace_data.get()[data["parent_trace_id"]].children.append(trace_id)


def trace(
    name: Optional[str] = None,
    tags: Optional[list[str]] = None,
    metadata: Optional[dict[str, Any]] = None,
    target: Optional[str] = None,
    end_user_identifier: Optional[str] = None,
    eval_funcs_names: Optional[list[str]] = None,
    eval_funcs: Optional[list[Callable]] = None,
    access_output_of_func: Optional[Callable] = None,
    apply_eval_frac: Optional[float] = 1.0,
):
    def init_trace(func_name, args, kwargs, func) -> tuple[str, float]:
        start_time = time.time()
        trace_id = gen_trace_id()
        trace_context.get().append(trace_id)

        sig = inspect.signature(func)
        parameters = sig.parameters

        inputs = {k: v for k, v in zip(parameters.keys(), args)}
        inputs.update(kwargs)

        # filter out any values which aren't JSON serializable
        for k, v in inputs.items():
            try:
                json.dumps(v)
            except TypeError:
                inputs[k] = json_dumps(v)

        trace_data.get()[trace_id] = TraceLog(
            trace_id=trace_id,
            parent_trace_id=trace_id,
            root_trace_id=trace_context.get()[0],
            start_timestamp=to_date_and_time_string(start_time),
            trace_name=name or func_name,
            end_user_identifier=end_user_identifier,
            metadata=metadata,
            target=target,
            tags=tags,
            inputs=inputs,
            experiment_uuid=os.environ.get(PAREA_OS_ENV_EXPERIMENT_UUID, None),
            apply_eval_frac=apply_eval_frac,
        )
        parent_trace_id = trace_context.get()[-2] if len(trace_context.get()) > 1 else None
        if parent_trace_id:
            fill_trace_data(trace_id, {"parent_trace_id": parent_trace_id}, UpdateTraceScenario.CHAIN)

        return trace_id, start_time

    def cleanup_trace(trace_id, start_time):
        end_time = time.time()
        trace_data.get()[trace_id].end_timestamp = to_date_and_time_string(end_time)
        trace_data.get()[trace_id].latency = end_time - start_time

        output = trace_data.get()[trace_id].output
        if trace_data.get()[trace_id].status == "success" and output:
            output_for_eval_metrics = output
            if access_output_of_func:
                try:
                    output = json.loads(output)
                    output = access_output_of_func(output)
                    output_for_eval_metrics = json_dumps(output)
                except Exception as e:
                    logger.exception(f"Error accessing output of func with output: {output}. Error: {e}", exc_info=e)
            trace_data.get()[trace_id].output_for_eval_metrics = output_for_eval_metrics

        thread_eval_funcs_then_log(trace_id, eval_funcs)
        trace_context.get().pop()

    def decorator(func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            trace_id, start_time = init_trace(func.__name__, args, kwargs, func)
            output_as_list = check_multiple_return_values(func)
            try:
                result = await func(*args, **kwargs)
                fill_trace_data(trace_id, {"result": result, "output_as_list": output_as_list, "eval_funcs_names": eval_funcs_names}, UpdateTraceScenario.RESULT)
            except Exception as e:
                logger.exception(f"Error occurred in function {func.__name__}, {e}")
                fill_trace_data(trace_id, {"error": str(e)}, UpdateTraceScenario.ERROR)
                raise e
            finally:
                cleanup_trace(trace_id, start_time)
            return result

        @wraps(func)
        def wrapper(*args, **kwargs):
            trace_id, start_time = init_trace(func.__name__, args, kwargs, func)
            output_as_list = check_multiple_return_values(func)
            try:
                result = func(*args, **kwargs)
                fill_trace_data(trace_id, {"result": result, "output_as_list": output_as_list, "eval_funcs_names": eval_funcs_names}, UpdateTraceScenario.RESULT)
            except Exception as e:
                logger.exception(f"Error occurred in function {func.__name__}, {e}")
                fill_trace_data(trace_id, {"error": str(e)}, UpdateTraceScenario.ERROR)
                raise e
            finally:
                cleanup_trace(trace_id, start_time)
            return result

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return wrapper

    if callable(name):
        func = name
        name = None
        return decorator(func)

    return decorator


def call_eval_funcs_then_log(trace_id: str, eval_funcs: list[Callable] = None):
    data = trace_data.get()[trace_id]
    parea_logger.default_log(data=data)

    if eval_funcs and data.status == "success" and random() <= data.apply_eval_frac:
        thread_ids_running_evals.get().append(trace_id)
        scores = []
        for func in eval_funcs:
            try:
                scores.append(NamedEvaluationScore(name=func.__name__, score=func(data)))
            except Exception as e:
                logger.exception(f"Error occurred calling evaluation function '{func.__name__}', {e}", exc_info=e)
        parea_logger.update_log(data=UpdateLog(trace_id=trace_id, field_name_to_value_map={"scores": scores}))
        thread_ids_running_evals.get().remove(trace_id)


def logger_record_log(trace_id: str):
    log_in_thread(parea_logger.record_log, {"data": trace_data.get()[trace_id]})


def logger_all_possible(trace_id: str):
    log_in_thread(parea_logger.default_log, {"data": trace_data.get()[trace_id]})


def thread_eval_funcs_then_log(trace_id: str, eval_funcs: list[Callable] = None):
    log_in_thread(call_eval_funcs_then_log, {"trace_id": trace_id, "eval_funcs": eval_funcs})



from typing import Any, Optional, Union

from enum import Enum

from attr import define


class TraceIntegrations(str, Enum):
    LANGCHAIN = "langchain"


class Role(str, Enum):
    user = "user"
    assistant = "assistant"
    system = "system"
    example_user = "example_user"
    example_assistant = "example_assistant"


@define
class Message:
    content: str
    role: Role = Role.user

    def to_dict(self) -> dict[str, str]:
        return {
            "content": self.content,
            "role": str(self.role),
        }


@define
class ModelParams:
    temp: float = 1.0
    top_p: float = 1.0
    frequency_penalty: float = 0.0
    presence_penalty: float = 0.0
    max_length: Optional[int] = None
    response_format: Optional[dict] = None


@define
class LLMInputs:
    """
    model choice should match an enabled model on the Parea platform.
    model options are:
    OPENAI_MODELS
        "gpt-3.5-turbo"
        "gpt-3.5-turbo-0301"
        "gpt-3.5-turbo-0613"
        "gpt-3.5-turbo-16k"
        "gpt-3.5-turbo-16k-0301"
        "gpt-3.5-turbo-16k-0613"
        "gpt-3.5-turbo-1106"
        "gpt-3.5-turbo-0125"
        "gpt-3.5-turbo-instruct"
        "gpt-4"
        "gpt-4-0314"
        "gpt-4-0613"
        "gpt-4-32k"
        "gpt-4-32k-0314"
        "gpt-4-32k-0613"
        "gpt-4-turbo-preview"
        "gpt-4-1106-preview"
        "gpt-4-0125-preview"
    You can use Azure models by providing your model name prefixed with azure_, e.g. azure_gpt-3.5-turbo
    ANTHROPIC_MODELS
        "claude-instant-1.1"
        "claude-instant-1"
        "claude-instant-1.2"
        "claude-instant-1-100k"
        "claude-instant-1.1-100k"
        "claude-1"
        "claude-2"
        "claude-1-100k"
        "claude-1.2"
        "claude-1.3"
        "claude-1.3-100k"
        "claude-2.1"
    AWS_ANTHROPIC_MODELS
        "anthropic.claude-instant-v1"
        "anthropic.claude-v1"
        "anthropic.claude-v2"
        "anthropic.claude-v2:1"
    ANYSCALE_MODELS
        "meta-llama/Llama-2-7b-chat-hf"
        "meta-llama/Llama-2-13b-chat-hf"
        "meta-llama/Llama-2-70b-chat-hf"
        "codellama/CodeLlama-34b-Instruct-hf"
        "mistralai/Mistral-7B-Instruct-v0.1"
        "mistralai/Mixtral-8x7B-Instruct-v0.1"
        "HuggingFaceH4/zephyr-7b-beta"
        "Open-Orca/Mistral-7B-OpenOrca"
    VERTEX_MODELS
        "gemini-pro"
        "text-bison@001"
        "text-bison@002"
        "text-bison"
        "text-bison-32k"
        "chat-bison@002"
        "chat-bison@001"
        "chat-bison"
        "chat-bison-32k
    """

    model: Optional[str] = None
    provider: Optional[str] = None
    model_params: Optional[ModelParams] = None
    messages: Optional[list[Message]] = None
    functions: Optional[list[Any]] = None
    function_call: Optional[Union[str, dict[str, str]]] = None


@define
class Log:
    configuration: LLMInputs = LLMInputs()
    inputs: Optional[dict[str, str]] = None
    output: Optional[str] = None
    target: Optional[str] = None
    latency: Optional[float] = 0.0
    input_tokens: Optional[int] = 0
    output_tokens: Optional[int] = 0
    total_tokens: Optional[int] = 0
    cost: Optional[float] = 0.0



from typing import Any, Callable, Optional, Union

import json
import os
from collections import defaultdict
from collections.abc import AsyncIterator, Iterator, Sequence

import openai
from openai import __version__ as openai_version

from ..utils.universal_encoder import json_dumps
from .utils import _calculate_input_tokens, _compute_cost, _format_function_call, _kwargs_to_llm_configuration, _num_tokens_from_string

if openai_version.startswith("0."):
    from openai.openai_object import OpenAIObject
    from openai.util import convert_to_openai_object
else:
    from openai.types.chat import ChatCompletion as OpenAIObject

    def convert_to_openai_object(kwargs) -> OpenAIObject:
        if "id" not in kwargs:
            kwargs["id"] = "0"
        if "created" not in kwargs:
            kwargs["created"] = 0
        if "object" not in kwargs:
            kwargs["object"] = "chat.completion"
        if "model" not in kwargs:
            kwargs["model"] = "model"
        if "choices" in kwargs and isinstance(kwargs["choices"], list) and len(kwargs["choices"]) > 0:
            if "finish_reason" not in kwargs["choices"][0]:
                kwargs["choices"][0]["finish_reason"] = "stop"
        return OpenAIObject(**kwargs)


from dotenv import load_dotenv

from ..cache.cache import Cache
from ..schemas.models import CacheRequest, TraceLog
from ..utils.trace_utils import trace_data
from .wrapper import Wrapper

load_dotenv()

is_old_openai = openai_version.startswith("0.")


class OpenAIWrapper:
    openai.api_key = os.getenv("OPENAI_API_KEY")

    def get_original_methods(self, module_client=openai):
        if is_old_openai:
            original_methods = {"ChatCompletion.create": module_client.ChatCompletion.create, "ChatCompletion.acreate": module_client.ChatCompletion.acreate}
        else:
            try:
                original_methods = {"chat.completions.create": module_client.chat.completions.create}
            except openai.OpenAIError:
                original_methods = {}
        return list(original_methods.keys())

    def init(self, log: Callable, cache: Cache = None, module_client=openai):
        Wrapper(
            resolver=self.resolver,
            gen_resolver=self.gen_resolver,
            agen_resolver=self.agen_resolver,
            should_use_gen_resolver=self.should_use_gen_resolver,
            log=log,
            module=module_client,
            func_names=self.get_original_methods(module_client),
            cache=cache,
            convert_kwargs_to_cache_request=self.convert_kwargs_to_cache_request,
            convert_cache_to_response=self.convert_cache_to_response,
            aconvert_cache_to_response=self.aconvert_cache_to_response,
        )

    def resolver(self, trace_id: str, _args: Sequence[Any], kwargs: dict[str, Any], response: Optional[Any]) -> Optional[Any]:
        if response:
            output = self._get_output(response)
            if openai_version.startswith("0."):
                usage = response["usage"]
                input_tokens = usage.get("prompt_tokens", 0)
                output_tokens = usage.get("completion_tokens", 0)
                total_tokens = usage.get("total_tokens", 0)
            else:
                usage = response.usage
                input_tokens = usage.prompt_tokens
                output_tokens = usage.completion_tokens
                total_tokens = usage.total_tokens
        else:
            output = None
            input_tokens = 0
            output_tokens = 0
            total_tokens = 0

        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        model = response.model if response and response.model else llm_configuration.model

        trace_data.get()[trace_id].configuration = llm_configuration
        trace_data.get()[trace_id].input_tokens = input_tokens
        trace_data.get()[trace_id].output_tokens = output_tokens
        trace_data.get()[trace_id].total_tokens = total_tokens
        trace_data.get()[trace_id].cost = _compute_cost(input_tokens, output_tokens, model)
        trace_data.get()[trace_id].output = output
        return response

    def gen_resolver(self, trace_id: str, _args: Sequence[Any], kwargs: dict[str, Any], response, final_log):
        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration

        accumulator, model_from_response = self._get_default_dict_streaming()
        for chunk in response:
            self._update_accumulator_streaming(accumulator, model_from_response, chunk)
            yield chunk

        self._format_accumulator_in_place(accumulator)

        model = model_from_response.get("model") or llm_configuration.model
        self.update_trace_data_from_stream_response(trace_id, model, accumulator, kwargs)
        final_log()

    @staticmethod
    def _get_default_dict_streaming():
        model_from_response = {"model": ""}
        accumulated_tools = defaultdict(lambda: {"id": "", "function": {"arguments": [], "name": ""}, "type": "function"})
        accumulator = defaultdict(content=[], role="assistant", function_call=defaultdict(arguments=[], name=""), tool_calls=accumulated_tools)
        return accumulator, model_from_response

    @staticmethod
    def _format_accumulator_in_place(accumulator):
        content = "".join(accumulator.get("content"))
        accumulator["content"] = content

        tool_calls = []
        if accumulated_tools := accumulator.get("tool_calls", {}):
            for tool in accumulated_tools.values():
                tool["function"]["arguments"] = "".join(tool["function"]["arguments"])

            tool_calls = list(accumulated_tools.values())
            for tool in tool_calls:
                tool["function"]["arguments"] = json_dumps(json.loads(tool["function"]["arguments"]), indent=2)
        accumulator["tool_calls"] = tool_calls

        if function_call := accumulator.get("function_call"):
            if function_call.get("name"):
                function_call["arguments"] = json_dumps(json.loads("".join(function_call["arguments"])), indent=2)
            else:
                function_call["arguments"] = ""

        return accumulator

    @staticmethod
    def _update_accumulator_streaming(accumulator, model_from_response, chunk):
        if chunk.model and not model_from_response.get("model"):
            model_from_response["model"] = chunk.model

        if not chunk.choices:
            return

        if is_old_openai:
            delta_dict = chunk.choices[0].delta._previous
        else:
            delta_dict = chunk.choices[0].delta.model_dump()

        if not accumulator["role"]:
            accumulator["role"] = delta_dict.get("role")

        if delta_dict.get("content"):
            accumulator["content"].append(delta_dict["content"])

        if delta_dict.get("function_call"):
            accumulator["function_call"]["name"] = delta_dict.get("function_call")["name"] or accumulator["function_call"]["name"]
            if delta_dict.get("function_call", {}).get("arguments"):
                accumulator["function_call"]["arguments"].append(delta_dict["function_call"]["arguments"])

        if tool_calls := delta_dict.get("tool_calls", []):
            for tool_call in tool_calls:
                tool_id = tool_call["index"]

                accumulator["tool_calls"][tool_id]["id"] = tool_call.get("id") or accumulator["tool_calls"][tool_id]["id"]
                if not accumulator["tool_calls"][tool_id]["type"]:
                    accumulator["tool_calls"][tool_id]["type"] = "function"

                accumulator["tool_calls"][tool_id]["function"]["name"] = tool_call.get("function", {}).get("name") or accumulator["tool_calls"][tool_id]["function"]["name"]
                if tool_call.get("function", {}).get("arguments"):
                    accumulator["tool_calls"][tool_id]["function"]["arguments"].append(tool_call["function"]["arguments"])

    async def agen_resolver(self, trace_id: str, _args: Sequence[Any], kwargs: dict[str, Any], response, final_log):
        llm_configuration = self._kwargs_to_llm_configuration(kwargs)
        trace_data.get()[trace_id].configuration = llm_configuration

        accumulator, model_from_response = self._get_default_dict_streaming()
        async for chunk in response:
            self._update_accumulator_streaming(accumulator, model_from_response, chunk)
            yield chunk

        self._format_accumulator_in_place(accumulator)

        model = model_from_response.get("model") or llm_configuration.model
        self.update_trace_data_from_stream_response(trace_id, model, accumulator, kwargs)
        final_log()

    def update_trace_data_from_stream_response(self, trace_id, model, accumulator, kwargs):
        output = self._get_output(accumulator, model)
        trace_data.get()[trace_id].output = output
        output_tokens = _num_tokens_from_string(output, model)
        input_tokens = _calculate_input_tokens(
            kwargs.get("messages", []),
            kwargs.get("functions", []) or [d["function"] for d in kwargs.get("tools", [])],
            kwargs.get("function_call", "auto") or kwargs.get("tool_choice", "auto"),
            model,
        )
        trace_data.get()[trace_id].input_tokens = input_tokens
        trace_data.get()[trace_id].output_tokens = output_tokens
        trace_data.get()[trace_id].total_tokens = input_tokens + output_tokens
        trace_data.get()[trace_id].cost = _compute_cost(input_tokens, output_tokens, model)

    @staticmethod
    def _kwargs_to_llm_configuration(kwargs):
        return _kwargs_to_llm_configuration(kwargs)

    def _get_output(self, result: Any, model: Optional[str] = None) -> str:
        if not isinstance(result, OpenAIObject) and isinstance(result, dict):
            result = convert_to_openai_object(
                {
                    "choices": [
                        {
                            "index": 0,
                            "message": result,
                        }
                    ],
                    "model": model or "model",
                }
            )
        response_message = result.choices[0].message
        if not response_message.get("content", None) if is_old_openai else not response_message.content:
            completion = self._format_function_call(response_message)
        else:
            completion = response_message.content.strip()
        return completion

    @staticmethod
    def _format_function_call(response_message) -> str:
        return _format_function_call(response_message)

    def convert_kwargs_to_cache_request(self, _args: Sequence[Any], kwargs: dict[str, Any]) -> CacheRequest:
        return CacheRequest(
            configuration=self._kwargs_to_llm_configuration(kwargs),
        )

    @staticmethod
    def _convert_cache_to_response(_args: Sequence[Any], kwargs: dict[str, Any], cache_response: TraceLog) -> OpenAIObject:
        content = cache_response.output
        message = {"role": "assistant", "content": None}
        try:
            function_or_tool_call = json.loads(content)
            if isinstance(function_or_tool_call, list) and len(function_or_tool_call) == 1:
                if "name" in function_or_tool_call[0] and "arguments" in function_or_tool_call[0]:
                    message["function_call"] = function_or_tool_call[0]
            elif isinstance(function_or_tool_call, list) and len(function_or_tool_call) > 1:
                tool_calls = []
                for idx, tool_call in enumerate(function_or_tool_call):
                    if "name" in tool_call and "arguments" in tool_call:
                        tool_calls.append({"id": idx, "function": tool_call, "type": "function"})
                message["tool_calls"] = tool_calls
            else:
                message["content"] = content
        except json.JSONDecodeError:
            message["content"] = content

        message_field = "delta" if kwargs.get("stream", False) else "message"

        return convert_to_openai_object(
            {
                "object": "chat.completion",
                "model": cache_response.configuration.model,
                "choices": [
                    {
                        "index": 0,
                        message_field: message,
                    }
                ],
                "usage": {
                    "prompt_tokens": cache_response.input_tokens,
                    "completion_tokens": cache_response.output_tokens,
                    "total_tokens": cache_response.total_tokens,
                },
            }
        )

    def convert_cache_to_response(self, _args: Sequence[Any], kwargs: dict[str, Any], cache_response: TraceLog) -> Union[OpenAIObject, Iterator[OpenAIObject]]:
        response = self._convert_cache_to_response(_args, kwargs, cache_response)
        if kwargs.get("stream", False):
            return iter([response])
        else:
            return response

    def aconvert_cache_to_response(self, _args: Sequence[Any], kwargs: dict[str, Any], cache_response: TraceLog) -> Union[OpenAIObject, AsyncIterator[OpenAIObject]]:
        response = self._convert_cache_to_response(_args, kwargs, cache_response)
        if kwargs.get("stream", False):

            def aiterator(iterable):
                async def gen():
                    for item in iterable:
                        yield item

                return gen()

            return aiterator([response])
        else:
            return response

    @staticmethod
    def should_use_gen_resolver(response: Any) -> bool:
        if is_old_openai:
            return isinstance(response, (Iterator, AsyncIterator))
        else:
            from openai import AsyncStream, Stream

            return isinstance(response, (Stream, AsyncStream))



import json
import os
import time

import openai
from attr import asdict
from dotenv import load_dotenv

from parea import InMemoryCache, Parea, get_current_trace_id, trace, write_trace_logs_to_csv
from parea.evals import call_openai
from parea.evals.chat import goal_success_ratio_factory
from parea.schemas import Log

load_dotenv()

openai.api_key = os.getenv("OPENAI_API_KEY")


use_cache = True  # by using the in memory cache, you don't need a Parea API key
cache = InMemoryCache() if use_cache else None
Parea(api_key=os.getenv("PAREA_API_KEY"), cache=cache)


def friendliness(log: Log) -> float:
    output = log.output
    response = call_openai(
        [
            {"role": "system", "content": "You evaluate the friendliness of the following response on a scale of 0 to 10. You must only return a number."},
            {"role": "assistant", "content": output},
        ],
        model="gpt-4",
    )
    try:
        return float(response) / 10.0
    except TypeError:
        return 0.0


def usefulness(log: Log) -> float:
    user_input = log.inputs["messages"][-1]["content"]
    output = log.output
    response = call_openai(
        [
            {"role": "system", "content": "You evaluate the usefulness of the response given the user input on a scale of 0 to 10. You must only return a number."},
            {"role": "assistant", "content": f'''User input: "{user_input}"\nAssistant response: "{output}"'''},
        ],
        model="gpt-4",
    )
    try:
        return float(response) / 10.0
    except TypeError:
        return 0.0


@trace(eval_funcs=[friendliness, usefulness])
def helpful_the_second_time(messages: list[dict[str, str]]) -> str:
    helpful_response = call_openai(
        [
            {"role": "system", "content": "You are a friendly, and helpful assistant that helps people with their homework."},
        ]
        + messages,
        model="gpt-4",
    )

    has_user_asked_before_raw = call_openai(
        [
            {
                "role": "system",
                "content": "Assess if the user has asked the last question before or is asking again for more \
information on a previous topic. If so, respond ASKED_BEFORE. Otherwise, respond NOT_ASKED_BEFORE.",
            }
        ]
        + messages,
        model="gpt-4",
    )
    has_user_asked_before = has_user_asked_before_raw == "ASKED_BEFORE"

    if has_user_asked_before:
        messages.append({"role": "assistant", "content": helpful_response})
        return helpful_response
    else:
        unhelfpul_response = call_openai(
            [
                {
                    "role": "system",
                    "content": "Given the helpful response to the user input below, please provide a slightly unhelpful \
    response which makes the user ask again in case they didn't ask already again because of a previous unhelpful answer. \
    In case the user asked again, please provide a last response",
                },
            ]
            + messages
            + [{"role": "assistant", "content": helpful_response}],
            model="gpt-4",
        )
        messages.append({"role": "assistant", "content": unhelfpul_response})
        return unhelfpul_response


goal_success_ratio = goal_success_ratio_factory(use_output=True)


@trace(eval_funcs=[goal_success_ratio], access_output_of_func=lambda x: x[0])
def unhelpful_chat():
    print("\nWelcome to the somewhat helpful chat! Type 'exit' to end the session.")

    trace_id = get_current_trace_id()

    messages = []
    while True:
        user_input = input("\nYou: ")

        if user_input.lower() == "exit":
            print("Goodbye!")
            break

        messages.append({"role": "user", "content": user_input})
        print("Bot:", helpful_the_second_time(messages))

    return messages, trace_id


def main():
    _, trace_id = unhelpful_chat()

    if os.getenv("PAREA_API_KEY"):
        print(f"You can view the logs at: https://app.parea.ai/logs/detailed/{trace_id}")
    if use_cache:
        time.sleep(5)  # wait for local eval function to finish
        path_csv = f"trace_logs-{int(time.time())}.csv"
        trace_logs = cache.read_logs()
        write_trace_logs_to_csv(path_csv, trace_logs)
        print(f"\nCSV-file of traces: {path_csv}")
        parent_trace = None
        for trace_log in trace_logs:
            if trace_log.trace_id == trace_id:
                parent_trace = trace_log
                break
        if parent_trace:
            print(f"Overall score(s):\n{json.dumps(parent_trace.scores, default=asdict, indent=2)}")


if __name__ == "__main__":
    main()



import os

import httpx
from dotenv import load_dotenv

from parea import Parea, aprocess_stream_and_yield, convert_openai_raw_to_log, process_stream_and_yield, trace
from parea.cookbook.data.openai_input_examples import functions_example, simple_example, tool_calling_example
from parea.wrapper import get_formatted_openai_response

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))

URL = "https://api.openai.com/v1/chat/completions"
HEADERS = {
    "Content-Type": "application/json",
    "Authorization": f"Bearer {os.getenv('OPENAI_API_KEY')}",
}
TIMEOUT = None

# Sync HTTPX


## Normal
@trace
def call_openai_api(data: dict) -> str:
    with httpx.Client(timeout=TIMEOUT) as client:
        response = client.post(URL, json=data, headers=HEADERS)
        r = response.json()
        convert_openai_raw_to_log(r, data)  # Add this line to enable tracing. Non-blocking
        return get_formatted_openai_response(r)  # Return how you normally would


## Streaming
@trace
def call_openai_api_stream(data: dict):
    data["stream"] = True
    with httpx.stream("POST", URL, json=data, headers=HEADERS, timeout=TIMEOUT) as response:
        # Add process_stream_and_yield to enable tracing. Non-blocking
        for chunk in process_stream_and_yield(response, data):
            print(chunk)


# Async HTTPX


## Normal
@trace
async def acall_openai_api(data: dict) -> str:
    async with httpx.AsyncClient(timeout=TIMEOUT) as client:
        response = await client.post(URL, json=data, headers=HEADERS)
        r = response.json()
        convert_openai_raw_to_log(r, data)  # Add this line to enable tracing. Non-blocking
        return get_formatted_openai_response(r)  # Return how you normally would


## Streaming
@trace
async def acall_openai_api_stream(data: dict):
    data["stream"] = True
    async with httpx.AsyncClient(timeout=TIMEOUT).stream("POST", URL, json=data, headers=HEADERS) as response:
        # Add process_stream_and_yield to enable tracing. Non-blocking
        async for chunk in aprocess_stream_and_yield(response, data):
            print(chunk)


# TEST NESTED TRACING
@trace
def chain():
    call_openai_api(simple_example)
    call_openai_api(functions_example)
    call_openai_api(tool_calling_example)
    call_openai_api_stream(tool_calling_example)


@trace
async def achain():
    await acall_openai_api(simple_example)
    await acall_openai_api(functions_example)
    await acall_openai_api(tool_calling_example)
    await acall_openai_api_stream(tool_calling_example)


if __name__ == "__main__":
    chain()
    # asyncio.run(achain())



from typing import Any

import dataclasses
import datetime
import json
from decimal import Decimal
from uuid import UUID

import attrs
from pydantic import BaseModel


def is_dataclass_instance(obj):
    return dataclasses.is_dataclass(obj) and not isinstance(obj, type)


def is_attrs_instance(obj):
    return attrs.has(obj)


def is_numpy_instance(obj):
    try:
        import numpy as np
    except ImportError:
        np = None

    return np and isinstance(obj, np.ndarray)


def is_pandas_instance(obj):
    try:
        import pandas as pd
    except ImportError:
        pd = None

    return pd and isinstance(obj, pd.DataFrame)


class UniversalEncoder(json.JSONEncoder):
    """
    A JSON encoder that can handle additional types such as dataclasses, attrs, and more.
    """

    def default(self, obj: Any):
        if isinstance(obj, str):
            return obj
        elif isinstance(obj, set):
            return list(obj)
        elif is_dataclass_instance(obj):
            return dataclasses.asdict(obj)
        elif is_attrs_instance(obj):
            return attrs.asdict(obj)
        elif isinstance(obj, BaseModel):
            try:
                return obj.model_dump()
            except Exception:
                return obj.dict()
        elif isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        elif isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, Decimal):
            return float(obj)
        elif callable(obj):
            return f"<callable {obj.__name__}>"
        elif isinstance(obj, bytes):
            return obj.decode(errors="ignore")
        elif is_numpy_instance(obj):
            return obj.tolist()
        elif is_pandas_instance(obj):
            return obj.to_dict(orient="records")
        else:
            return super().default(obj)


def json_dumps(obj, **kwargs):
    return json.dumps(obj, cls=UniversalEncoder, **kwargs)



from typing import Union

from uuid import UUID

from langchain_core.tracers import BaseTracer
from langchain_core.tracers.schemas import ChainRun, LLMRun, Run, ToolRun

from parea.parea_logger import parea_logger
from parea.schemas.log import TraceIntegrations


class PareaAILangchainTracer(BaseTracer):
    parent_trace_id: UUID

    def _persist_run(self, run: Union[Run, LLMRun, ChainRun, ToolRun]) -> None:
        self.parent_trace_id = run.id
        # using .dict() since langchain Run class currently set to Pydantic v1
        data = run.dict()
        parea_logger.record_vendor_log(data, TraceIntegrations.LANGCHAIN)

    def get_parent_trace_id(self) -> UUID:
        return self.parent_trace_id



import os

from dotenv import load_dotenv
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI

from parea import Parea, PareaAILangchainTracer

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))
handler = PareaAILangchainTracer()

llm = ChatOpenAI(openai_api_key=os.getenv("OPENAI_API_KEY"))
prompt = ChatPromptTemplate.from_messages([("user", "{input}")])
chain = prompt | llm | StrOutputParser()


def main():
    return chain.invoke(
        {"input": "Write a Hello World program in Python using FastAPI."},
        config={"callbacks": [handler]},
    )


if __name__ == "__main__":
    print(main())



from typing import Optional

import csv
import random
import time
import uuid

from attr import asdict, fields_dict

from parea.constants import ADJECTIVES, NOUNS
from parea.schemas.models import TraceLog


def gen_trace_id() -> str:
    """Generate a unique trace id for each chain of requests"""
    return str(uuid.uuid4())


def to_date_and_time_string(timestamp: float) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S %Z", time.localtime(timestamp))


def date_and_time_string_to_timestamp(date_and_time_string: str) -> float:
    return time.mktime(time.strptime(date_and_time_string, "%Y-%m-%d %H:%M:%S %Z"))


def write_trace_logs_to_csv(path_csv: str, trace_logs: list[TraceLog]):
    with open(path_csv, "w", newline="") as file:
        # write header
        columns = fields_dict(TraceLog).keys()
        writer = csv.DictWriter(file, fieldnames=columns)
        writer.writeheader()
        # write rows
        for trace_log in trace_logs:
            writer.writerow(asdict(trace_log))


def gen_random_name():
    random_generator = random.Random()
    adjective = random_generator.choice(ADJECTIVES)
    noun = random_generator.choice(NOUNS)
    return f"{adjective}-{noun}"


def calculate_avg_as_string(values: list[Optional[float]]) -> str:
    if not values:
        return "N/A"
    values = [x for x in values if x is not None]
    avg = sum(values) / len(values)
    return f"{avg:.2f}"



import os

from dotenv import load_dotenv
from langchain.chains import create_extraction_chain
from langchain.schema import HumanMessage
from langchain_experimental.llms.anthropic_functions import AnthropicFunctions

from parea import Parea, PareaAILangchainTracer

load_dotenv()

p = Parea(api_key=os.getenv("PAREA_API_KEY"))

model = AnthropicFunctions(model="claude-2")

functions = [
    {
        "name": "get_current_weather",
        "description": "Get the current weather in a given location",
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. San Francisco, CA",
                },
                "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
            },
            "required": ["location"],
        },
    }
]


schema = {
    "properties": {
        "name": {"type": "string"},
        "height": {"type": "integer"},
        "hair_color": {"type": "string"},
    },
    "required": ["name", "height"],
}
inp = """Alex is 5 feet tall. Claudia is 1 feet taller Alex and jumps higher than him. Claudia is a brunette and Alex
is blonde."""

chain = create_extraction_chain(schema, model)


def main():
    response = model.predict_messages([HumanMessage(content="whats the weater in boston?")], functions=functions, callbacks=[PareaAILangchainTracer()])
    print(response)
    result = chain.run(inp, callbacks=[PareaAILangchainTracer()])
    print(result)


if __name__ == "__main__":
    main()



from typing import Any, Callable, Optional, Union

import asyncio
import os
import time
from collections.abc import AsyncIterable, Iterable

import httpx
from attrs import asdict, define, field
from cattrs import structure
from dotenv import load_dotenv

from parea.api_client import HTTPClient
from parea.cache import InMemoryCache, RedisCache
from parea.cache.cache import Cache
from parea.constants import PAREA_OS_ENV_EXPERIMENT_UUID
from parea.experiment.datasets import create_test_collection
from parea.helpers import gen_trace_id
from parea.parea_logger import parea_logger
from parea.schemas.models import (
    Completion,
    CompletionResponse,
    CreateExperimentRequest,
    CreateGetProjectResponseSchema,
    CreateTestCaseCollection,
    ExperimentSchema,
    ExperimentStatsSchema,
    FeedbackRequest,
    ProjectSchema,
    TestCaseCollection,
    UseDeployedPrompt,
    UseDeployedPromptResponse,
)
from parea.utils.trace_utils import get_current_trace_id, get_root_trace_id, logger_all_possible, logger_record_log, trace_data
from parea.wrapper import OpenAIWrapper

load_dotenv()

COMPLETION_ENDPOINT = "/completion"
DEPLOYED_PROMPT_ENDPOINT = "/deployed-prompt"
RECORD_FEEDBACK_ENDPOINT = "/feedback"
EXPERIMENT_ENDPOINT = "/experiment"
EXPERIMENT_STATS_ENDPOINT = "/experiment/{experiment_uuid}/stats"
EXPERIMENT_FINISHED_ENDPOINT = "/experiment/{experiment_uuid}/finished"
PROJECT_ENDPOINT = "/project"
GET_COLLECTION_ENDPOINT = "/collection/{test_collection_name}"
CREATE_COLLECTION_ENDPOINT = "/collection"


@define
class Parea:
    api_key: str = field(init=True, default=os.getenv("PAREA_API_KEY"))
    project_name: str = field(init=True, default="default")
    cache: Cache = field(init=True, default=None)
    _project: ProjectSchema = field(init=False, default=None)
    _client: HTTPClient = field(init=False, default=HTTPClient())

    def __attrs_post_init__(self):
        self._client.set_api_key(self.api_key)
        project_api_response: CreateGetProjectResponseSchema = self._create_or_get_project(self.project_name)
        if project_api_response.was_created:
            print(f"Created project {project_api_response.name}")
        self._project = structure(asdict(project_api_response), ProjectSchema)

        if self.api_key:
            parea_logger.set_client(self._client)
            parea_logger.set_project_uuid(self.project_uuid)
        if isinstance(self.cache, (RedisCache, InMemoryCache)):
            parea_logger.set_redis_cache(self.cache)
        _init_parea_wrapper(logger_all_possible, self.cache)

    def wrap_openai_client(self, client: "OpenAI") -> None:
        """Only necessary for instance client with OpenAI version >= 1.0.0"""
        OpenAIWrapper().init(log=logger_all_possible, cache=self.cache, module_client=client)

    def _add_project_uuid_to_data(self, data) -> dict:
        data_dict = asdict(data)
        data_dict["project_uuid"] = self._project.uuid
        return data_dict

    def completion(self, data: Completion) -> CompletionResponse:
        parent_trace_id = get_current_trace_id()
        inference_id = gen_trace_id()
        data.inference_id = inference_id
        data.parent_trace_id = parent_trace_id or inference_id
        data.root_trace_id = get_root_trace_id()
        data.project_uuid = self._project.uuid

        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data.experiment_uuid = experiment_uuid

        r = self._client.request(
            "POST",
            COMPLETION_ENDPOINT,
            data=asdict(data),
        )

        if parent_trace_id:
            trace_data.get()[parent_trace_id].children.append(inference_id)
            trace_data.get()[parent_trace_id].experiment_uuid = experiment_uuid
            logger_record_log(parent_trace_id)

        return structure(r.json(), CompletionResponse)

    async def acompletion(self, data: Completion) -> CompletionResponse:
        parent_trace_id = get_current_trace_id()
        inference_id = gen_trace_id()
        data.inference_id = inference_id
        data.parent_trace_id = parent_trace_id or inference_id
        data.root_trace_id = get_root_trace_id()
        data.project_uuid = self._project.uuid

        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data.experiment_uuid = experiment_uuid

        r = await self._client.request_async(
            "POST",
            COMPLETION_ENDPOINT,
            data=asdict(data),
        )

        if parent_trace_id:
            trace_data.get()[parent_trace_id].children.append(inference_id)
            trace_data.get()[parent_trace_id].experiment_uuid = experiment_uuid
            logger_record_log(parent_trace_id)

        return structure(r.json(), CompletionResponse)

    def stream(self, data: Completion) -> Iterable[bytes]:
        parent_trace_id = get_current_trace_id()
        inference_id = gen_trace_id()
        data.inference_id = inference_id
        data.parent_trace_id = parent_trace_id or inference_id
        data.root_trace_id = get_root_trace_id()
        data.project_uuid = self._project.uuid

        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data.experiment_uuid = experiment_uuid

        if parent_trace_id:
            trace_data.get()[parent_trace_id].children.append(inference_id)
            trace_data.get()[parent_trace_id].experiment_uuid = experiment_uuid
            logger_record_log(parent_trace_id)

        response = self._client.stream_request(
            "POST",
            f"{COMPLETION_ENDPOINT}/stream",
            data=asdict(data),
        )
        yield from response

    async def astream(self, data: Completion) -> AsyncIterable[bytes]:
        parent_trace_id = get_current_trace_id()
        inference_id = gen_trace_id()
        data.inference_id = inference_id
        data.parent_trace_id = parent_trace_id or inference_id
        data.root_trace_id = get_root_trace_id()
        data.project_uuid = self._project.uuid

        if experiment_uuid := os.getenv(PAREA_OS_ENV_EXPERIMENT_UUID, None):
            data.experiment_uuid = experiment_uuid

        if parent_trace_id:
            trace_data.get()[parent_trace_id].children.append(inference_id)
            trace_data.get()[parent_trace_id].experiment_uuid = experiment_uuid
            logger_record_log(parent_trace_id)

        response = self._client.stream_request_async(
            "POST",
            f"{COMPLETION_ENDPOINT}/stream",
            data=asdict(data),
        )
        async for chunk in response:
            yield chunk

    def get_prompt(self, data: UseDeployedPrompt) -> UseDeployedPromptResponse:
        r = self._client.request(
            "POST",
            DEPLOYED_PROMPT_ENDPOINT,
            data=asdict(data),
        )
        return structure(r.json(), UseDeployedPromptResponse)

    async def aget_prompt(self, data: UseDeployedPrompt) -> UseDeployedPromptResponse:
        r = await self._client.request_async(
            "POST",
            DEPLOYED_PROMPT_ENDPOINT,
            data=asdict(data),
        )
        return structure(r.json(), UseDeployedPromptResponse)

    def record_feedback(self, data: FeedbackRequest) -> None:
        time.sleep(2)  # give logs time to update
        self._client.request(
            "POST",
            RECORD_FEEDBACK_ENDPOINT,
            data=asdict(data),
        )

    async def arecord_feedback(self, data: FeedbackRequest) -> None:
        await asyncio.sleep(2)  # give logs time to update
        await self._client.request_async(
            "POST",
            RECORD_FEEDBACK_ENDPOINT,
            data=asdict(data),
        )

    def create_experiment(self, data: CreateExperimentRequest) -> ExperimentSchema:
        try:
            r = self._client.request(
                "POST",
                EXPERIMENT_ENDPOINT,
                data=self._add_project_uuid_to_data(data),
            )
            return structure(r.json(), ExperimentSchema)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 400:
                response_detail = e.response.json()
                error_msg = response_detail.get("detail", e.response.text)
                raise ValueError(f"Error creating experiment: {error_msg}")
            raise

    async def acreate_experiment(self, data: CreateExperimentRequest) -> ExperimentSchema:
        try:
            r = await self._client.request_async(
                "POST",
                EXPERIMENT_ENDPOINT,
                data=self._add_project_uuid_to_data(data),
            )
            return structure(r.json(), ExperimentSchema)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 400:
                response_detail = e.response.json()
                error_msg = response_detail.get("detail", e.response.text)
                raise ValueError(f"Error creating experiment: {error_msg}")
            raise

    def get_experiment_stats(self, experiment_uuid: str) -> ExperimentStatsSchema:
        r = self._client.request(
            "GET",
            EXPERIMENT_STATS_ENDPOINT.format(experiment_uuid=experiment_uuid),
        )
        return structure(r.json(), ExperimentStatsSchema)

    async def aget_experiment_stats(self, experiment_uuid: str) -> ExperimentStatsSchema:
        r = await self._client.request_async(
            "GET",
            EXPERIMENT_STATS_ENDPOINT.format(experiment_uuid=experiment_uuid),
        )
        return structure(r.json(), ExperimentStatsSchema)

    def finish_experiment(self, experiment_uuid: str) -> ExperimentStatsSchema:
        r = self._client.request(
            "POST",
            EXPERIMENT_FINISHED_ENDPOINT.format(experiment_uuid=experiment_uuid),
        )
        return structure(r.json(), ExperimentStatsSchema)

    async def afinish_experiment(self, experiment_uuid: str) -> ExperimentStatsSchema:
        r = await self._client.request_async(
            "POST",
            EXPERIMENT_FINISHED_ENDPOINT.format(experiment_uuid=experiment_uuid),
        )
        return structure(r.json(), ExperimentStatsSchema)

    def _create_or_get_project(self, name: str) -> CreateGetProjectResponseSchema:
        r = self._client.request(
            "POST",
            PROJECT_ENDPOINT,
            data=asdict(CreateExperimentRequest(name=name)),
        )
        return structure(r.json(), CreateGetProjectResponseSchema)

    def get_collection(self, test_collection_name: str) -> TestCaseCollection:
        r = self._client.request(
            "GET",
            GET_COLLECTION_ENDPOINT.format(test_collection_name=test_collection_name),
        )
        return structure(r.json(), TestCaseCollection)

    async def aget_collection(self, test_collection_name: str) -> TestCaseCollection:
        r = await self._client.request_async(
            "GET",
            GET_COLLECTION_ENDPOINT.format(test_collection_name=test_collection_name),
        )
        return structure(r.json(), TestCaseCollection)

    def create_test_collection(self, data: list[dict[str, Any]], name: Optional[str] = None) -> None:
        request: CreateTestCaseCollection = create_test_collection(data, name)
        self._client.request(
            "POST",
            CREATE_COLLECTION_ENDPOINT,
            data=asdict(request),
        )

    @property
    def project_uuid(self) -> str:
        return self._project.uuid

    def experiment(self, data: Union[str, Iterable[dict]], func: Callable):
        """
        :param data: If your dataset is defined locally it should be an iterable of k/v
        pairs matching the expected inputs of your function. To reference a test collection you
        have saved on Parea, use the collection name as a string.

        :param func: The function to run. This function should accept inputs that match the keys of the data field.
        """
        from parea import Experiment

        return Experiment(data=data, func=func, p=self)


_initialized_parea_wrapper = False


def _init_parea_wrapper(log: Callable = None, cache: Cache = None):
    global _initialized_parea_wrapper
    if _initialized_parea_wrapper:
        return
    OpenAIWrapper().init(log=log, cache=cache)
    _initialized_parea_wrapper = True
